# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import cirq


class QSimCircuit(cirq.Circuit):

  def __init__(self,
               cirq_circuit: cirq.Circuit,
               device: cirq.devices = cirq.devices.UNCONSTRAINED_DEVICE,
               allow_decomposition: bool = False):

    if allow_decomposition:
      super().__init__([], device)
      for moment in cirq_circuit:
        for op in moment:
          # This should call decompose on the gates
          self.append(op)
    else:
      super().__init__(cirq_circuit, device)

  def __eq__(self, other):
    if not isinstance(other, QSimCircuit):
      return False
    # equality is tested, for the moment, for cirq.Circuit
    return super().__eq__(other)

  def _resolve_parameters_(self, param_resolver: cirq.study.ParamResolver):

    qsim_circuit = super()._resolve_parameters_(param_resolver)

    qsim_circuit.device = self.device

    return qsim_circuit

  def translate_cirq_to_qsim(
      self,
      qubit_order: cirq.ops.QubitOrderOrList = cirq.ops.QubitOrder.DEFAULT
  ) -> str:
    """
        Translates this Cirq circuit to the qsim representation
        :qubit_order: Ordering of qubits
        :return: the string representing line-by-line the qsim circuit
        """

    circuit_data = []
    ordered_qubits = cirq.ops.QubitOrder.as_qubit_order(qubit_order).order_for(
        self.all_qubits())
    qubit_to_index_dict = {q: i for i, q in enumerate(ordered_qubits)}
    for mi, moment in enumerate(self):
      for op in moment:

        qub_str = ""
        for qub in op.qubits:
          qub_str += "{} ".format(qubit_to_index_dict[qub])

        qsim_gate = ""
        qsim_params = ""
        if isinstance(op.gate, cirq.ops.HPowGate)\
                and op.gate.exponent == 1.0:
          qsim_gate = "h"
        elif isinstance(op.gate, cirq.ops.ZPowGate) \
                and op.gate.exponent == 0.25:
          qsim_gate = "t"
        elif isinstance(op.gate, cirq.ops.XPowGate) \
                and op.gate.exponent == 1.0:
          qsim_gate = "x"
        elif isinstance(op.gate, cirq.ops.YPowGate) \
                and op.gate.exponent == 1.0:
          qsim_gate = "y"
        elif isinstance(op.gate, cirq.ops.ZPowGate) \
                and op.gate.exponent == 1.0:
          qsim_gate = "z"
        elif isinstance(op.gate, cirq.ops.XPowGate) \
                and op.gate.exponent == 0.5:
          qsim_gate = "x_1_2"
        elif isinstance(op.gate, cirq.ops.YPowGate) \
                and op.gate.exponent == 0.5:
          qsim_gate = "y_1_2"
        elif isinstance(op.gate, cirq.ops.XPowGate):
          qsim_gate = "rx"
          qsim_params = str(op.gate.exponent)
        elif isinstance(op.gate, cirq.ops.YPowGate):
          qsim_gate = "ry"
          qsim_params = str(op.gate.exponent)
        elif isinstance(op.gate, cirq.ops.ZPowGate):
          qsim_gate = "rz"
          qsim_params = str(op.gate.exponent)
        elif isinstance(op.gate, cirq.ops.CZPowGate) \
                and op.gate.exponent == 1.0:
          qsim_gate = "cz"
        elif isinstance(op.gate, cirq.ops.CNotPowGate) \
                and op.gate.exponent == 1.0:
          qsim_gate = "cnot"
        elif isinstance(op.gate, cirq.ops.ISwapPowGate) \
                and op.gate.exponent == 1.0:
          qsim_gate = "is"
        elif isinstance(op.gate, cirq.ops.FSimGate):
          qsim_gate = "fs"
          qsim_params = "{} {}".format(op.gate.theta, op.gate.phi)
        else:
          raise ValueError("{!r} No translation for ".format(op))

        # The moment is missing
        qsim_gate = "{} {} {} {}".format(mi, qsim_gate, qub_str.strip(),
                                         qsim_params.strip())
        circuit_data.append(qsim_gate.strip())

    circuit_data.insert(0, str(len(ordered_qubits)))
    return "\n".join(circuit_data)



import os
import re
import sys
import platform
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion


class CMakeExtension(Extension):

  def __init__(self, name, sourcedir=''):
    Extension.__init__(self, name, sources=[])
    self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):

  def run(self):
    try:
      out = subprocess.check_output(['cmake', '--version'])
    except OSError:
      raise RuntimeError(
          'CMake must be installed to build the following extensions: ' +
          ', '.join(e.name for e in self.extensions))

    if platform.system() == 'Windows':
      cmake_version = LooseVersion(
          re.search(r'version\s*([\d.]+)', out.decode()).group(1))
      if cmake_version < '3.1.0':
        raise RuntimeError('CMake >= 3.1.0 is required on Windows')

    for ext in self.extensions:
      self.build_extension(ext)

  def build_extension(self, ext):
    extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
    cmake_args = [
        '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
        '-DPYTHON_EXECUTABLE=' + sys.executable
    ]

    cfg = 'Debug' if self.debug else 'Release'
    build_args = ['--config', cfg]

    if platform.system() == 'Windows':
      cmake_args += [
          '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)
      ]
      if sys.maxsize > 2**32:
        cmake_args += ['-A', 'x64']
      build_args += ['--', '/m']
    else:
      cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
      build_args += ['--', '-j2']

    env = os.environ.copy()
    env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
        env.get('CXXFLAGS', ''), self.distribution.get_version())
    if not os.path.exists(self.build_temp):
      os.makedirs(self.build_temp)
    subprocess.check_call(
        ['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
    subprocess.check_call(
        ['cmake', '--build', '.'] + build_args, cwd=self.build_temp)


requirements = open('requirements.txt').readlines()

description = ('Schrödinger and Schrödinger-Feynman simulators for quantum circuits.')

# README file as long_description.
long_description = open('README.md', encoding='utf-8').read()

setup(
    name='qsimcirq',
    version='0.0.2',
    author='Vamsi Krishna Devabathini',
    author_email='devabathini92@gmail.com',
    python_requires=('>=3.3.0'),
    install_requires=requirements,
    license='Apache 2',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    ext_modules=[CMakeExtension('qsimcirq/qsim')],
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
    packages=['qsimcirq'])
